

#include "ArticulatedObjectExampleMemoryWriterClient.h"

#include <memory>

#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/VirtualRobot.h>

#include <ArmarXCore/core/PackagePath.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/armem/core/Time.h>


namespace armarx::articulated_object
{
    ArticulatedObjectExampleMemoryWriterClient::ArticulatedObjectExampleMemoryWriterClient() :
        articulatedObjectWriter(new ::armarx::armem::articulated_object::Writer(*this)) {}

    armarx::PropertyDefinitionsPtr ArticulatedObjectExampleMemoryWriterClient::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        // defs->optional(memoryName, "mem.MemoryName", "Name of the memory to use.");

        articulatedObjectWriter->registerPropertyDefinitions(defs);

        return defs;
    }

    std::string ArticulatedObjectExampleMemoryWriterClient::getDefaultName() const
    {
        return "ArticulatedObjectExampleMemoryWriterClient";
    }

    void ArticulatedObjectExampleMemoryWriterClient::onInitComponent() {}

    void ArticulatedObjectExampleMemoryWriterClient::onConnectComponent()
    {
        articulatedObjectWriter->connect();

        task = new RunningTask<ArticulatedObjectExampleMemoryWriterClient>(this, &ArticulatedObjectExampleMemoryWriterClient::run);
        task->start();
    }

    void ArticulatedObjectExampleMemoryWriterClient::onDisconnectComponent()
    {
        task->stop();
    }

    void ArticulatedObjectExampleMemoryWriterClient::onExitComponent() {}

    VirtualRobot::RobotPtr createDishwasher()
    {
        const std::string xml =
            "./ArmarXObjects/Environment/mobile-kitchen/dishwasher-only/dishwasher.xml";

        return VirtualRobot::RobotIO::loadRobot(ArmarXDataPath::resolvePath(xml), VirtualRobot::RobotIO::eStructure);
    }

    armem::articulated_object::ArticulatedObject convert(const VirtualRobot::Robot& obj, const armem::Time& timestamp)
    {
        ARMARX_INFO << "Filename is " << obj.getFilename();

        return
            armem::articulated_object::ArticulatedObject
        {
            .description = {
                .name = obj.getName(),
                .xml = PackagePath(armarx::ArmarXDataPath::getProject({"ArmarXObjects"}, obj.getFilename()), obj.getFilename())
            },
            .instance = "", // TODO(fabian.reister):
            .config = {
                .timestamp = timestamp,
                .globalPose = Eigen::Affine3f(obj.getRootNode()->getGlobalPose()),
                .jointMap = obj.getJointValues()
            },
            .timestamp = timestamp
        };
    }

    void ArticulatedObjectExampleMemoryWriterClient::run()
    {
        ARMARX_IMPORTANT << "Running example.";

        std::shared_ptr<VirtualRobot::Robot> dishwasher = createDishwasher();

        CycleUtil cycle(IceUtil::Time::milliSeconds(100));
        IceUtil::Time start = TimeUtil::GetTime();

        CycleUtil c(100);

        while (not task->isStopped())
        {
            ARMARX_INFO << "Reporting articulated objects";

            const IceUtil::Time now = TimeUtil::GetTime();
            const float t     = float((now - start).toSecondsDouble());

            // move joints at certain frequency
            const float k = (1 + std::sin(t / (M_2_PIf32))) / 2; // in [0,1]

            const std::map<std::string, float> jointValues
            {
                {"dishwasher_door_joint", M_PIf32 / 2 * k},
                {"drawer_joint", 350 * k}
            };

            dishwasher->setJointValues(jointValues);

            armarx::armem::articulated_object::ArticulatedObject armemDishwasher = convert(*dishwasher, IceUtil::Time::now());
            articulatedObjectWriter->store(armemDishwasher);

            c.waitForCycleDuration();
        }
    }

} // namespace armarx::articulated_object
