/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ExampleMemoryClient
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ExampleMemoryClient.h"

#include <SimoxUtility/color/cmaps.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>

#include <RobotAPI/components/armem/server/ExampleMemory/aron/ExampleData.aron.generated.h>


namespace armarx
{
    ExampleMemoryClientPropertyDefinitions::ExampleMemoryClientPropertyDefinitions(std::string prefix) :
        armarx::ComponentPropertyDefinitions(prefix)
    {
    }

    armarx::PropertyDefinitionsPtr ExampleMemoryClient::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ExampleMemoryClientPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(memoryName, "mem.MemoryName", "Name of the memory to use.");

        return defs;
    }

    std::string ExampleMemoryClient::getDefaultName() const
    {
        return "ExampleMemoryClient";
    }


    void ExampleMemoryClient::onInitComponent()
    {
    }


    void ExampleMemoryClient::onConnectComponent()
    {
        createRemoteGuiTab();
        RemoteGui_startRunningTask();

        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "Waiting for memory '" << memoryName << "' ...";
        auto result = useMemory(memoryName);
        if (!result.success)
        {
            ARMARX_ERROR << result.errorMessage;
        }

        // Add a provider segment to commit to.
        providerID = addProviderSegment();
        // Construct the entity ID.
        entityID = providerID.withEntityName("example_entity");


        // Subscribe to example_entity updates
        // Using a lambda:
        memoryReader.subscribe(entityID, [&](const armem::MemoryID & entityID, const std::vector<armem::MemoryID>& snapshotIDs)
        {
            ARMARX_INFO << "Entity " << entityID << " was updated by " << snapshotIDs.size() << " snapshots.";
        });
        // Using a member function:
        memoryReader.subscribe(entityID, this, &ExampleMemoryClient::example_entityUpdated);


        task = new RunningTask<ExampleMemoryClient>(this, &ExampleMemoryClient::run);
        task->start();
    }


    void ExampleMemoryClient::onDisconnectComponent()
    {
        task->stop();
    }


    void ExampleMemoryClient::onExitComponent()
    {
    }


    void ExampleMemoryClient::run()
    {
        ARMARX_IMPORTANT << "Running example.";

        armem::MemoryID snapshotID = commitSingleSnapshot(entityID);
        if (true)
        {
            commitMultipleSnapshots(entityID, 3);
        }
        if (true)
        {
            queryLatestSnapshot(snapshotID.getEntityID());
        }
        if (true)
        {
            queryExactSnapshot(snapshotID);
        }
        if (true)
        {
            commitExampleData();
            queryExampleData();
        }

        CycleUtil c(1000);
        int i = 0;
        while (!task->isStopped() && i++ < 100)
        {
            commitSingleSnapshot(entityID);
            c.waitForCycleDuration();
        }
    }


    armem::MemoryID ExampleMemoryClient::addProviderSegment()
    {
        armem::data::AddSegmentInput input;
        input.coreSegmentName = "ExampleModality";
        input.providerSegmentName = "FancyMethodModality";

        ARMARX_IMPORTANT << input;
        armem::data::AddSegmentResult result = memoryWriter.addSegment(input);
        ARMARX_INFO << result;

        return armem::MemoryID(result.segmentID);
    }

    armem::MemoryID ExampleMemoryClient::commitSingleSnapshot(const armem::MemoryID& entityID)
    {
        // Prepare the update with some empty instances.
        armem::EntityUpdate update;
        update.entityID = entityID;
        update.timeCreated = armem::Time::now();
        update.instancesData =
        {
            std::make_shared<aron::datanavigator::DictNavigator>(),
            std::make_shared<aron::datanavigator::DictNavigator>()
        };

        ARMARX_IMPORTANT << "Committing " << update;
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);
        ARMARX_INFO << updateResult;
        if (!updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
        }

        return updateResult.snapshotID;
    }

    void ExampleMemoryClient::commitMultipleSnapshots(const armem::MemoryID& entityID, int num)
    {
        // Commit a number of updates with different timestamps and number of instances.
        armem::Commit commit;
        for (int i = 0; i < num; ++i)
        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = entityID;
            update.timeCreated = armem::Time::now() + armem::Time::seconds(i);
            for (int j = 0; j < i; ++j)
            {
                update.instancesData.push_back(std::make_shared<aron::datanavigator::DictNavigator>());
            }
        }
        ARMARX_IMPORTANT << "Committing " << commit;
        armem::CommitResult commitResult = memoryWriter.commit(commit);
        ARMARX_INFO << commitResult;
        if (!commitResult.allSuccess())
        {
            ARMARX_ERROR << commitResult.allErrorMessages();
        }
    }


    void ExampleMemoryClient::queryLatestSnapshot(const armem::MemoryID& entityID)
    {
        ARMARX_IMPORTANT
                << "Querying latest snapshot: "
                << "\n- entityID:     \t'" << entityID << "'"
                ;

        armem::client::query::Builder builder;
        builder
        .coreSegments().withID(entityID)
        .providerSegments().withID(entityID)
        .entities().withID(entityID)
        .snapshots().latest();

        armem::client::QueryResult qResult = memoryReader.query(builder.buildQueryInput());
        ARMARX_INFO << qResult;

        if (qResult.success)
        {
            armem::wm::Memory& memory = qResult.memory;

            ARMARX_CHECK_EQUAL(memory.coreSegments().size(), 1);
            armem::wm::CoreSegment& coreSeg = memory.coreSegments().begin()->second;

            ARMARX_CHECK_EQUAL(coreSeg.providerSegments().size(), 1);
            armem::wm::ProviderSegment& provSeg = coreSeg.providerSegments().begin()->second;

            ARMARX_CHECK_EQUAL(provSeg.entities().size(), 1);
            armem::wm::Entity& entity = provSeg.entities().begin()->second;

            ARMARX_CHECK_GREATER_EQUAL(entity.history().size(), 1);
            armem::wm::EntitySnapshot& snapshot = entity.history().begin()->second;

            ARMARX_INFO << "Result: "
                        << "\n- memory:       \t" << memory.name()
                        << "\n- core segment: \t" << coreSeg.name()
                        << "\n- prov segment: \t" << provSeg.name()
                        << "\n- entity:       \t" << entity.name()
                        << "\n- snapshot:     \t" << snapshot.time()
                        << "\n- #instances:   \t" << snapshot.size()
                        ;

            ARMARX_IMPORTANT << "Getting entity via ID";
            const armem::wm::Entity& e = memory.getEntity(entityID);
            ARMARX_INFO << "Result: "
                        << "\n- entity:       \t" << e.name()
                        << "\n- snapshot:     \t" << e.getLatestSnapshot().time()
                        << "\n- #instances:   \t" << e.getLatestSnapshot().size()
                        ;

            // Show memory contents in remote gui.
            tab.queryResult = std::move(memory);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }
    }

    void ExampleMemoryClient::queryExactSnapshot(const armem::MemoryID& snapshotID)
    {
        ARMARX_IMPORTANT
                << "Querying exact snapshot: "
                << "\n- snapshotID:     \t'" << snapshotID << "'"
                ;

        namespace qf = armem::client::query_fns;
        armem::client::query::Builder qb;
        qb
        .coreSegments(qf::withID(snapshotID))
        .providerSegments(qf::withID(snapshotID))
        .entities(qf::withID(snapshotID))
        .snapshots(qf::withID(snapshotID));

        armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());
        ARMARX_INFO << qResult;

        if (qResult.success)
        {
            armem::wm::Memory memory = std::move(qResult.memory);
            const armem::wm::EntitySnapshot& entitySnapshot = memory.getEntity(snapshotID).getLatestSnapshot();

            ARMARX_INFO << "Result snapshot: "
                        << "\n- time:        \t" << entitySnapshot.time()
                        << "\n- # instances: \t" << entitySnapshot.size()
                        ;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }
    }


    void ExampleMemoryClient::commitExampleData()
    {
        ARMARX_IMPORTANT << "Adding segment " << "ExampleData" << "/" << getName();

        auto addSegmentResult = memoryWriter.addSegment("ExampleData", getName());
        if (!addSegmentResult.success)
        {
            ARMARX_ERROR << addSegmentResult.errorMessage;
            return;
        }
        armem::MemoryID providerID = armem::MemoryID(addSegmentResult.segmentID);

        armem::Time time = armem::Time::now();
        armem::Commit commit;
        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID.withEntityName("default");
            update.timeCreated = time;

            armem::example::ExampleData primitive;
            update.instancesData = { primitive.toAron() };
        }
        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID.withEntityName("the answer");
            update.timeCreated = time;

            armem::example::ExampleData data;
            data.the_bool = true;
            data.the_double = 2 * 42.42;
            data.the_float = 21.5;
            data.the_int = 42;
            data.the_long = 424242;
            data.the_string = "fourty two";
            data.the_float_list = { 21, 42, 84 };
            data.the_int_list = { 21, 42, 84 };
            data.the_string_list = simox::alg::multi_to_string(data.the_int_list);
            data.the_object_list.emplace_back();

            data.the_float_dict =
            {
                { "one", 1.0 },
                { "two", 2.0 },
                { "three", 3.0 },
            };
            data.the_int_dict =
            {
                { "one", 1 },
                { "two", 2 },
                { "three", 3 },
            };
            {
                CByteImage& image = *data.the_ivt_image;
                image.Set(20, 10, CByteImage::ImageType::eRGB24);
                simox::ColorMap cmap = simox::color::cmaps::plasma();
                cmap.set_vmax(image.width + image.height);
                for (int row = 0; row < image.height; ++row)
                {
                    for (int col = 0; col < image.width; ++col)
                    {
                        simox::Color color = cmap(row + col);
                        unsigned char* p = &image.pixels[(row * image.width + col) * image.bytesPerPixel];
                        p[0] = color.r;
                        p[1] = color.g;
                        p[2] = color.b;
                    }
                }
            }

            update.instancesData = { data.toAron() };
        }

        armem::CommitResult commitResult = memoryWriter.commit(commit);
        if (!commitResult.allSuccess())
        {
            ARMARX_WARNING << commitResult.allErrorMessages();
        }
    }

    void ExampleMemoryClient::queryExampleData()
    {
        // Query all entities from provider.
        armem::client::query::Builder qb;
        qb
        .coreSegments().withID(providerID)
        .providerSegments().withID(providerID)
        .entities().all()
        .snapshots().all();

        armem::client::QueryResult result = memoryReader.query(qb.buildQueryInput());
        if (result)
        {
            tab.queryResult = std::move(result.memory);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << result.errorMessage;
        }
    }



    void ExampleMemoryClient::example_entityUpdated(const armem::MemoryID& subscriptionID, const std::vector<armem::MemoryID>& snapshotIDs)
    {
        std::stringstream ss;
        ss << "example_entity got updated: " << subscriptionID << "\n";
        ss << "Updated snapshots: \n";
        for (const auto& id : snapshotIDs)
        {
            ss << "- " << id << "\n";
        }
        ARMARX_IMPORTANT << ss.str();
        // Fetch new data of example_entity and do something with it.
    }


    void ExampleMemoryClient::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        if (tab.queryResult)
        {
            tab.queryResultGroup = armem::server::MemoryRemoteGui().makeGroupBox(*tab.queryResult);
        }

        VBoxLayout root = {tab.queryResultGroup, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }

    void ExampleMemoryClient::RemoteGui_update()
    {
        if (tab.rebuild.exchange(false))
        {
            createRemoteGuiTab();
        }
    }

}
