#include "ClientPlugin.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"



namespace armarx::armem::mns::plugins
{

    ClientPlugin::~ClientPlugin()
    {}


    void ClientPlugin::postCreatePropertyDefinitions(armarx::PropertyDefinitionsPtr& properties)
    {
        if (!properties->hasDefinition(makePropertyName(PROPERTY_MNS_NAME_NAME)))
        {
            properties->defineOptionalProperty<std::string>(
                makePropertyName(PROPERTY_MNS_NAME_NAME),
                parent<ClientPluginUserBase>().memoryNameSystemName,
                "Name of the Memory Name System (MNS) component.");
        }
        if (!properties->hasDefinition(makePropertyName(PROPERTY_MNS_ENABLED_NAME)))
        {
            properties->defineOptionalProperty<bool>(
                makePropertyName(PROPERTY_MNS_ENABLED_NAME),
                parent<ClientPluginUserBase>().memoryNameSystemEnabled,
                "Whether to use (and depend on) the Memory Name System (MNS)."
                "\nSet to false to use this memory as a stand-alone.");
        }
    }

    void ClientPlugin::preOnInitComponent()
    {
        if (isMemoryNameSystemEnabled())
        {
            parent().usingProxy(getMemoryNameSystemName());
        }
    }

    void ClientPlugin::preOnConnectComponent()
    {
        if (isMemoryNameSystemEnabled())
        {
            parent<ClientPluginUserBase>().memoryNameSystem = getMemoryNameSystem();
        }
    }

    bool ClientPlugin::isMemoryNameSystemEnabled()
    {
        return parentDerives<Component>() ?
               parent<Component>().getProperty<bool>(makePropertyName(PROPERTY_MNS_ENABLED_NAME)) :
               parent<ClientPluginUserBase>().memoryNameSystemEnabled;
    }

    std::string ClientPlugin::getMemoryNameSystemName()
    {
        return parentDerives<Component>() ?
               parent<Component>().getProperty<std::string>(makePropertyName(PROPERTY_MNS_NAME_NAME)) :
               std::string{parent<ClientPluginUserBase>().memoryNameSystemName};
    }

    mns::MemoryNameSystemInterfacePrx ClientPlugin::getMemoryNameSystem()
    {
        return isMemoryNameSystemEnabled() && parentDerives<Component>()
               ? parent<Component>().getProxy<MemoryNameSystemInterfacePrx>(getMemoryNameSystemName())
               : nullptr;
    }

    ClientPluginUserBase::~ClientPluginUserBase()
    {
    }

    armem::data::WaitForMemoryResult ClientPluginUserBase::useMemory(const MemoryID& id)
    {
        return useMemory(id.memoryName);
    }

    armem::data::WaitForMemoryResult ClientPluginUserBase::useMemory(const std::string& memoryName)
    {
        armem::data::WaitForMemoryInput input;
        input.name = memoryName;

        armem::data::WaitForMemoryResult result = memoryNameSystem->waitForMemory(input);
        if (result.success)
        {
            if (Component* comp = dynamic_cast<Component*>(this))
            {
                // Add dependency.
                comp->usingProxy(result.proxy->ice_getIdentity().name);
            }
        }
        return result;
    }

    armem::data::WaitForMemoryResult ClientPluginUserBase::waitForMemory(const std::string& memoryName)
    {
        if (!memoryNameSystem)
        {
            return {};
        }
        armem::data::WaitForMemoryInput input;
        input.name = memoryName;
        return memoryNameSystem->waitForMemory(input);
    }

    armem::data::ResolveMemoryNameResult ClientPluginUserBase::resolveMemoryName(const std::string& memoryName)
    {
        if (!memoryNameSystem)
        {
            return {};
        }
        armem::data::ResolveMemoryNameInput input;
        input.name = memoryName;
        return memoryNameSystem->resolveMemoryName(input);
    }

    bool ClientPluginUserBase::isMemoryAvailable(const std::string& memoryName)
    {
        armem::data::ResolveMemoryNameResult result = resolveMemoryName(memoryName);
        return result.success && result.proxy;
    }

}

