#pragma once

#include <RobotAPI/interface/armem/query.h>

#include "BaseQueryProcessorBase.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>


namespace armarx::armem::base::query_proc
{

    /**
     * @brief Handles memory queries.
     */
    template <class _EntityT>
    class EntityQueryProcessorBase :
        virtual public BaseQueryProcessorBase<_EntityT, armem::query::data::EntityQuery>
    {
        using Base = BaseQueryProcessorBase<_EntityT, armem::query::data::EntityQuery>;

    public:
        using EntityT = _EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;

        using Base::process;
        void process(_EntityT& result,
                     const armem::query::data::EntityQuery& query,
                     const _EntityT& entity) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::entity::All*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::Single*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::TimeRange*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::IndexRange*>(&query))
            {
                process(result, *q, entity);
            }
            else
            {
                throw armem::error::UnknownQueryType("entity snapshot", query);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::All& query,
                     const _EntityT& entity) const
        {
            (void) query;
            // Copy this entitiy and its contents.

            for (const auto& [time, snapshot] : entity)
            {
                addResultSnapshot(result, snapshot);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::Single& query,
                     const _EntityT& entity) const
        {
            if (query.timestamp < 0)
            {
                try
                {
                    addResultSnapshot(result, entity.getLatestSnapshot());
                }
                catch (const armem::error::EntityHistoryEmpty&)
                {
                    // Leave empty.
                }
            }
            else
            {
                Time time = fromIce<Time>(query.timestamp);
                try
                {
                    addResultSnapshot(result, entity.getSnapshot(time));
                }
                catch (const armem::error::MissingEntry&)
                {
                    // Leave empty.
                }
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::TimeRange& query,
                     const _EntityT& entity) const
        {
            if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
            {
                Time min = fromIce<Time>(query.minTimestamp);
                Time max = fromIce<Time>(query.maxTimestamp);
                process(result, min, max, entity, query);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::IndexRange& query,
                     const _EntityT& entity) const
        {
            if (entity.empty())
            {
                return;
            }

            size_t first = negativeIndexSemantics(query.first, entity.history().size());
            size_t last = negativeIndexSemantics(query.last, entity.history().size());

            if (first <= last)
            {
                auto it = entity.begin();
                std::advance(it, first);

                size_t num = last - first + 1;  // +1 to make last inclusive
                for (size_t i = 0; i < num; ++i, ++it)
                {
                    addResultSnapshot(result, it);
                }
            }
        }

        void process(_EntityT& result,
                     const Time& min,
                     const Time& max,
                     const _EntityT& entity,
                     const armem::query::data::EntityQuery& query) const
        {
            (void) query;

            // Returns an iterator pointing to the first element that is not less than (i.e. greater or equal to) key.
            auto begin = min.toMicroSeconds() > 0 ? entity.history().lower_bound(min) : entity.history().begin();
            // Returns an iterator pointing to the first element that is *greater than* key.
            auto end = max.toMicroSeconds() > 0 ? entity.history().upper_bound(max) : entity.history().end();

            for (auto it = begin; it != end && it != entity.history().end(); ++it)
            {
                addResultSnapshot(result, it);
            }
        }


        static size_t negativeIndexSemantics(long index, size_t size)
        {
            const size_t max = size > 0 ? size - 1 : 0;
            if (index >= 0)
            {
                return std::clamp<size_t>(static_cast<size_t>(index), 0, max);
            }
            else
            {
                return static_cast<size_t>(std::clamp<long>(static_cast<long>(size) + index, 0, static_cast<long>(max)));
            }
        }

    protected:
        virtual void addResultSnapshot(_EntityT& result, typename _EntityT::ContainerT::const_iterator it) const = 0;
        virtual void addResultSnapshot(_EntityT& result, const typename _EntityT::EntitySnapshotT& snapshot) const = 0;

    };
}
