/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "TransformWriter.h"

#include <algorithm>
#include <iterator>
#include <numeric>

#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>
#include <RobotAPI/libraries/armem/core/workingmemory/Memory.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ProviderSegment.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/aron/core/navigator/type/NavigatorFactory.h>
#include <RobotAPI/libraries/core/FramedPose.h>

#include <RobotAPI/libraries/armem_robot_localization/aron/Transform.aron.generated.h>

#include "aron_conversions.h"


namespace armarx::armem
{

    TransformWriter::TransformWriter(ManagedIceObject& component) : MemoryConnector(component) {}

    TransformWriter::~TransformWriter() = default;

    void TransformWriter::registerPropertyDefinitions(armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "TransformWriter: registerPropertyDefinitions";

        MemoryConnector::registerPropertyDefinitions(def);

        const std::string prefix = getPropertyPrefix();

        def->optional(properties.localizationMemoryName,
                      prefix + "LocalizationMemoryName",
                      "Name of the localization memory core segment to use.");

        def->optional(properties.memoryName, prefix + "MemoryName");
    }

    void TransformWriter::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "TransformWriter: Waiting for memory '" << properties.memoryName
                         << "' ...";
        auto result = useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "TransformWriter: Connected to memory '" << properties.memoryName;

        memoryWriter.setWritingMemory(result.proxy);
    }

    bool TransformWriter::commitTransform(const Transform& transform)
    {
        std::lock_guard g{memoryWriterMutex};

        ARMARX_DEBUG << "Trying to create core segment + provider segment";

        const auto result =
            memoryWriter.addSegment(properties.localizationMemoryName, transform.header.agent);

        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;

            // TODO(fabian.reister): message
            return false;
        }

        const auto timestamp = IceUtil::Time::microSeconds(transform.header.timestamp);

        const auto providerId = armem::MemoryID(result.segmentID);
        const auto entityID   = providerId.withEntityName(transform.header.parentFrame + "," +
                                transform.header.frame).withTimestamp(timestamp);

        armem::EntityUpdate update;
        update.entityID    = entityID;

        aron::Transform aronTransform;
        toAron(transform, aronTransform);

        update.instancesData = {aronTransform.toAron()};
        update.timeCreated = IceUtil::Time::microSeconds(aronTransform.header.timestamp);

        ARMARX_DEBUG << "Committing " << update << " at time " << IceUtil::Time::microSeconds(transform.header.timestamp);
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);

        ARMARX_DEBUG << updateResult;

        if (not updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
        }

        return updateResult.success;
    }

    const std::string& TransformWriter::getPropertyPrefix() const
    {
        return propertyPrefix;
    }


} // namespace armarx::armem
