/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <numeric>

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx::aron
{
    class Randomizer
    {
    public:
        Randomizer()
        {
            initialize_random();
        };

        typenavigator::NavigatorPtr generateRandomType(bool mustBeObject = false) const
        {
            type::Descriptor nextType = type::Descriptor::eObject;
            if (!mustBeObject)
            {
                std::vector<type::Descriptor> descriptors =
                {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    type::Descriptor::e##upperType,

                    HANDLE_ALL_ARON_TYPES
#undef RUN_ARON_MACRO
                };

                int randomType = generateRandom(descriptors.size(), 0);
                nextType = descriptors[randomType];
            }

            switch (nextType)
            {
                case type::Descriptor::eObject:
                {
                    typenavigator::ObjectNavigatorPtr t = typenavigator::ObjectNavigatorPtr(new typenavigator::ObjectNavigator(Path()));
                    std::string objectName = generateRandomWord();
                    t->setObjectName(objectName);

                    std::vector<std::string> usedNames({objectName});
                    int members = generateRandom(4, 2);
                    std::set<std::string> usedKeys;
                    for (int i = 0; i < members; ++i)
                    {
                        std::string key = generateRandomWord(usedKeys);
                        usedKeys.insert(key);

                        typenavigator::NavigatorPtr m = generateRandomType(false);
                        t->addMemberType(key, m);
                    }
                    return t;
                }
                case type::Descriptor::eDict:
                {
                    typenavigator::DictNavigatorPtr t = typenavigator::DictNavigatorPtr(new typenavigator::DictNavigator());
                    typenavigator::NavigatorPtr a = generateRandomType(false);

                    t->setAcceptedType(a);
                    return t;
                }
                case type::Descriptor::eTuple:
                {
                    typenavigator::TupleNavigatorPtr t = typenavigator::TupleNavigatorPtr(new typenavigator::TupleNavigator());

                    int members = generateRandom(4, 2);
                    for (int i = 0; i < members; ++i)
                    {
                        typenavigator::NavigatorPtr m = generateRandomType(false);
                        t->addAcceptedType(m);
                    }
                    return t;
                }
                case type::Descriptor::eList:
                {
                    typenavigator::ListNavigatorPtr t = typenavigator::ListNavigatorPtr(new typenavigator::ListNavigator());
                    typenavigator::NavigatorPtr a = generateRandomType(false);

                    t->setAcceptedType(a);
                    return t;
                }
                case type::Descriptor::ePair:
                {
                    typenavigator::PairNavigatorPtr t = typenavigator::PairNavigatorPtr(new typenavigator::PairNavigator());
                    typenavigator::NavigatorPtr a = generateRandomType(false);
                    typenavigator::NavigatorPtr b = generateRandomType(false);

                    t->setFirstAcceptedType(a);
                    t->setSecondAcceptedType(b);
                    return t;
                }
                case type::Descriptor::eEigenMatrix:
                case type::Descriptor::eEigenQuaternion:
                case type::Descriptor::eIVTCByteImage:
                case type::Descriptor::eOpenCVMat:
                case type::Descriptor::ePCLPointCloud:
                case type::Descriptor::ePosition:
                case type::Descriptor::eOrientation:
                case type::Descriptor::ePose:

                case type::Descriptor::eInt:
                {
                    auto t = std::make_shared<typenavigator::IntNavigator>();
                    return t;
                }
                case type::Descriptor::eLong:
                {
                    auto t = std::make_shared<typenavigator::LongNavigator>();
                    return t;
                }
                case type::Descriptor::eFloat:
                {
                    auto t = std::make_shared<typenavigator::FloatNavigator>();
                    return t;
                }
                case type::Descriptor::eDouble:
                {
                    auto t = std::make_shared<typenavigator::DoubleNavigator>();
                    return t;
                }
                case type::Descriptor::eString:
                {
                    auto t = std::make_shared<typenavigator::StringNavigator>();
                    return t;
                }
                case type::Descriptor::eBool:
                {
                    auto t = std::make_shared<typenavigator::BoolNavigator>();
                    return t;
                }
                case type::Descriptor::eTime:
                {
                    auto t = std::make_shared<typenavigator::TimeNavigator>();
                    return t;
                }
                default:
                {
                    throw error::DescriptorNotValidException("Randomizer", "generateRandomType", "No valid type found!", nextType);
                }
            }
        }

        datanavigator::NavigatorPtr generateAronDataFromType(const typenavigator::NavigatorPtr& type) const
        {
            const type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {
                // In an object, we do not want to edit the keys.
                case type::Descriptor::eObject:
                {
                    typenavigator::ObjectNavigatorPtr t = typenavigator::ObjectNavigator::DynamicCastAndCheck(type);
                    datanavigator::DictNavigatorPtr d = datanavigator::DictNavigatorPtr(new datanavigator::DictNavigator());
                    for (const auto& [k, tt] : t->getMemberTypes())
                    {
                        d->addElement(k, generateAronDataFromType(tt));
                    }
                    return d;
                }

                // here all totally random
                case type::Descriptor::eDict:
                {
                    typenavigator::DictNavigatorPtr t = typenavigator::DictNavigator::DynamicCastAndCheck(type);
                    return datanavigator::NavigatorPtr(new datanavigator::DictNavigator());
                }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case type::Descriptor::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    return datanavigator::NavigatorPtr(new datanavigator::ListNavigator()); \
}

                HANDLE_LIST_SERIALIZER_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case type::Descriptor::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    datanavigator::NDArrayNavigatorPtr ndarray = datanavigator::NDArrayNavigatorPtr(new datanavigator::NDArrayNavigator()); \
    return ndarray; \
}

                HANDLE_NDARRAY_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType, upperData, lowerData, capsData) \
case type::Descriptor::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    return datanavigator::NavigatorPtr(new datanavigator::upperData##Navigator()); \
}

                HANDLE_PRIMITIVE_CORRESPONDING
#undef RUN_ARON_MACRO
                default:
                {
                    throw error::DescriptorNotValidException("Randomizer", "generateAronDataFromType", "No valid type found!", desc);
                }
            }
        }

        void initializeRandomly(datanavigator::NavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            // Containers
            type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {

#define RUN_ARON_MACRO(upperType, lowerType, capsType, upperData, lowerData, capsData) \
case type::Descriptor::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    datanavigator::upperData##NavigatorPtr d = datanavigator::upperData##Navigator::DynamicCastAndCheck(data); \
    initializeRandomly(d, t); \
    break; \
}

                HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO

                default:
                {
                    throw error::DescriptorNotValidException("Randomizer", "initializeRandomly", "No valid type found!", desc);
                }
            }
        }

        int generateRandom(int max, int min) const
        {
            max += 1;
            int random = (std::rand() % (max - min)) + min;
            return random;
        }

        std::string generateRandomWord(const std::set<std::string>& usedKeys = {}) const
        {
            std::vector<string> words =
            {
                "Lorem", "ipsum", "dolor", "sit", "amet", "consetetur", "sadipscing", "elitr"
                "sed", "diam", "nonumy", "eirmod", "tempor", "invidunt", "ut", "labore", "et"
                "dolore", "magna", "aliquyam", "eratsed"
            };

            int i = generateRandom(words.size() - 1, 0);
            std::string key = words[i];

            while (usedKeys.count(key) > 0)
            {
                key = generateRandomWord(usedKeys);
            }

            return key;
        }

        std::vector<unsigned char> generateRandomBlob(unsigned int size) const
        {
            std::vector<unsigned char> new_blob(size, 0);
            for (unsigned int i = 0; i < size; ++i)
            {
                new_blob[i] = (generateRandom(127, 0));
            }
            return new_blob;
        }

    private:
        void initialize_random() const
        {
            std::srand(std::time(nullptr));
        }

    public:
        void initializeRandomly(datanavigator::DictNavigatorPtr& data, const typenavigator::ObjectNavigatorPtr& type) const
        {
            for (auto& [key, nextData] : data->getElements())
            {
                initializeRandomly(nextData, type->getMemberTypes()[key]);
            }
        }

        void initializeRandomly(datanavigator::DictNavigatorPtr& data, const typenavigator::DictNavigatorPtr& type) const
        {
            data->clear();
            int numElements = generateRandom(5, 1);
            std::set<std::string> usedKeys;
            for (int i = 0; i < numElements; ++i)
            {
                std::string key = generateRandomWord(usedKeys);
                usedKeys.insert(key);
                datanavigator::NavigatorPtr newData = generateAronDataFromType(type->getAcceptedType());
                initializeRandomly(newData, type->getAcceptedType());
                data->addElement(key, newData);
            }
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::TupleNavigatorPtr& type) const
        {
            unsigned int i = 0;
            for (auto& nextData : data->getElements())
            {
                initializeRandomly(nextData, type->getAcceptedType(i++));
            }
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::PairNavigatorPtr& type) const
        {
            datanavigator::NavigatorPtr first = data->getElement(0);
            datanavigator::NavigatorPtr second = data->getElement(1);
            initializeRandomly(first, type->getFirstAcceptedType());
            initializeRandomly(second, type->getSecondAcceptedType());
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::ListNavigatorPtr& type) const
        {
            data->clear();
            int numElements = generateRandom(5, 1);
            for (int i = 0; i < numElements; ++i)
            {
                datanavigator::NavigatorPtr newData = generateAronDataFromType(type->getAcceptedType());
                initializeRandomly(newData, type->getAcceptedType());
                data->addElement(newData);
            }
        }

        void initializeRandomly(datanavigator::NDArrayNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            std::vector<int> dims = data->getDimensions();
            int bytes = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
            std::vector<unsigned char> blob = generateRandomBlob(bytes);
            data->setData(bytes, blob.data());
        }

        void initializeRandomly(datanavigator::IntNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::LongNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::FloatNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::DoubleNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::BoolNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1, 0));
        }

        void initializeRandomly(datanavigator::StringNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandomWord());
        }
    };
}
