/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/Primitive.h>



namespace armarx::aron::dataIO::reader
{
    NavigatorReader::NavigatorReader(const datanavigator::NavigatorPtr& n) :
        input(n)
    {
    }

    NavigatorReader::NavigatorReader(const data::AronDataPtr& n) :
        input(datanavigator::Navigator::FromAronData(n))
    {
    }

    datanavigator::NavigatorPtr NavigatorReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        auto lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    datanavigator::NavigatorPtr NavigatorReader::getNext()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        auto lastToken = stack.top();
        return lastToken->getNextElement();
    }

    // Containers
    int NavigatorReader::readStartDict()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = datanavigator::DictNavigator::DynamicCastAndCheck(current_nav);
        int c = current_nav->childrenSize();
        auto newToken = std::make_shared<NavigatorReaderToken>(data::Descriptor::eDict, current_nav_casted);
        stack.push(newToken);
        return c;
    }

    bool NavigatorReader::readEndDict()
    {
        auto token = stack.top();
        datanavigator::DictNavigator::DynamicCastAndCheck(token->getElement());

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NavigatorReader::readStartList()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = datanavigator::ListNavigator::DynamicCastAndCheck(current_nav);
        int c = current_nav->childrenSize();
        auto newToken = std::make_shared<NavigatorReaderToken>(data::Descriptor::eList, current_nav_casted);
        stack.push(newToken);
        return c;
    }

    bool NavigatorReader::readEndList()
    {
        auto token = stack.top();
        datanavigator::ListNavigator::DynamicCastAndCheck(token->getElement());

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    // Complex Data
    std::tuple<std::vector<int>, std::string> NavigatorReader::readStartNDArray()
    {
        datanavigator::NavigatorPtr nav = getNext();
        datanavigator::NDArrayNavigatorPtr casted = datanavigator::NDArrayNavigator::DynamicCastAndCheck(nav);
        return {casted->getDimensions(), casted->getType()};
    }

    bool NavigatorReader::readEndNDArray(unsigned char* data)
    {
        if (data == NULL)
        {
            throw error::AronException("NavigatorReader", "readEndNDArray", "The corresponding data pointer of an complex aron ptr is NULL.");
        }
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::NDArrayNavigator::DynamicCastAndCheck(nav);
        std::vector<int> dims = casted->getDimensions();
        memcpy(data, casted->getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
        return true;
    }

    // Read primitives
    void NavigatorReader::readPrimitive(int& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::IntNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    void NavigatorReader::readPrimitive(long& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::LongNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    void NavigatorReader::readPrimitive(float& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::FloatNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    void NavigatorReader::readPrimitive(double& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::DoubleNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    void NavigatorReader::readPrimitive(std::string& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::StringNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    void NavigatorReader::readPrimitive(bool& t)
    {
        auto nav = getNextAndIncrease();
        auto casted = datanavigator::BoolNavigator::DynamicCastAndCheck(nav);
        t = casted->getValue();
    }

    std::string NavigatorReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NavigatorReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    data::Descriptor NavigatorReader::getTypeOfNext(const type::Descriptor hint) const
    {
        if (stack.empty())
        {
            if (hint == type::Descriptor::eUnknown)
            {
                return data::Descriptor::eDict;
            }
            else
            {
                return Resolver::GetCorresponding(hint);
            }
        }
        auto token = stack.top();
        return token->getTypeOfNext(hint);
    }
}
