/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Simox
#include <SimoxUtility/json.h>

// Base Class
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/ReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/Resolver.h>



namespace armarx::aron::dataIO::reader
{
    // BaseClass
    class NlohmannJSONReaderToken;
    typedef std::shared_ptr<NlohmannJSONReaderToken> NlohmannJSONReaderTokenPtr;

    class NlohmannJSONReaderToken :
        virtual public ReaderTokenInterface<nlohmann::json>
    {
    public:
        using PointerType = NlohmannJSONReaderTokenPtr;

        // constructors
        NlohmannJSONReaderToken() = delete;
        NlohmannJSONReaderToken(const data::Descriptor desc, const nlohmann::json& data)
        {
            descriptor = desc;
            element = data;

            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    for (auto it = data.begin(); it != data.end(); ++it)
                    {
                        allMemberNames.push_back(it.key());
                    }
                    childrenSize = data.size();
                    break;
                }
                case data::Descriptor::eList:
                {
                    childrenSize = data.size();
                    break;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "NlohmannJSONReaderToken", "Received an invalid instance for a reader token.", desc);
            }
        }

        data::Descriptor getTypeOfNext(const type::Descriptor hint) const override
        {
            const nlohmann::json next = getNextElement();
            if (next.is_object())
            {
                // Check if specific NDArray key exists
                if (next.find(io::Data::READER_WRITER_NDARRAY_DATA_SLUG) != next.end())
                {
                    return data::Descriptor::eNDArray;
                }
                return data::Descriptor::eDict;
            }
            if (next.is_array())
            {
                return data::Descriptor::eList;
            }
            if (next.is_number_integer())
            {

                data::Descriptor d = Resolver::GetFirstIfRelated(
                                         Resolver::GetCorresponding(hint),
                                         data::Descriptor::eLong);
                return d;
            }
            if (next.is_number_float())
            {
                data::Descriptor d = Resolver::GetFirstIfRelated(
                                         Resolver::GetCorresponding(hint),
                                         data::Descriptor::eDouble);
                return d;
            }
            if (next.is_boolean())
            {
                return data::Descriptor::eBool;
            }
            if (next.is_string())
            {
                return data::Descriptor::eString;
            }
            throw error::AronException("NlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Could not convert to data::Descriptor enum.");
        }

        nlohmann::json getNextElement() const override
        {
            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    nlohmann::json ret = element[getCurrentKey()];
                    return ret;
                }
                case data::Descriptor::eList:
                {
                    nlohmann::json ret = element[currentIndex];
                    return ret;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
            }
        }

    private:
    };
}
