/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorWriter.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/navigator/data/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/Primitive.h>

#include <RobotAPI/libraries/aron/core/navigator/type/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Tuple.h>
#include <RobotAPI/libraries/aron/core/navigator/type/primitive/Primitive.h>

namespace armarx::aron::typeIO::writer
{
    // generate Path info
    Path NavigatorWriter::generatePath()
    {
        Path path;
        if (!wroteInitialStartObject)
        {
            wroteInitialStartObject = true;
        }
        else
        {
            auto current = stack.top();
            path = Path(current->getElement()->getPath(), current->toAccessor());
        }
        return path;
    }

    // interface
    void NavigatorWriter::writeStartObject(const std::string& n)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ObjectNavigator>(path);
        type->setObjectName(n);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndObject()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartDict()
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::DictNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartList()
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ListNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartTuple()
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::TupleNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndTuple()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartPair()
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::PairNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndPair()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeEigenMatrix(const std::vector<int>& dims, const std::string& t)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::EigenMatrixNavigatorPtr aron(new typenavigator::EigenMatrixNavigator(path));
        aron->setTypename(t);
        aron->setDimensions(dims);
        token->addElement(aron);
    }

    void NavigatorWriter::writeEigenQuaternion(const std::string& t)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::EigenQuaternionNavigatorPtr aron(new typenavigator::EigenQuaternionNavigator(path));
        aron->setTypename(t);
        token->addElement(aron);
    }

    void NavigatorWriter::writeIVTCByteImage(unsigned int w, unsigned int h, const std::string& t)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::IVTCByteImageNavigatorPtr aron(new typenavigator::IVTCByteImageNavigator(path));
        aron->setTypename(t);
        aron->setWidth(w);
        aron->setHeight(h);
        token->addElement(aron);
    }

    void NavigatorWriter::writeOpenCVMat(const std::vector<int>& dims, const std::string& t)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::OpenCVMatNavigatorPtr aron(new typenavigator::OpenCVMatNavigator(path));
        aron->setTypename(t);
        aron->setDimensions(dims);
        token->addElement(aron);
    }

    void NavigatorWriter::writePCLPointCloud(unsigned int w, unsigned int h, const std::string& t)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::PCLPointCloudNavigatorPtr aron(new typenavigator::PCLPointCloudNavigator(path));
        aron->setTypename(t);
        aron->setWidth(w);
        aron->setHeight(h);
        token->addElement(aron);
    }

    void NavigatorWriter::writePosition()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::PositionNavigatorPtr aron(new typenavigator::PositionNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeOrientation()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::OrientationNavigatorPtr aron(new typenavigator::OrientationNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writePose()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::PoseNavigatorPtr aron(new typenavigator::PoseNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeInt()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::IntNavigatorPtr aron(new typenavigator::IntNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeLong()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::LongNavigatorPtr aron(new typenavigator::LongNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeFloat()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::FloatNavigatorPtr aron(new typenavigator::FloatNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeDouble()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::DoubleNavigatorPtr aron(new typenavigator::DoubleNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeString()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::StringNavigatorPtr aron(new typenavigator::StringNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeBool()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::BoolNavigatorPtr aron(new typenavigator::BoolNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeTime()
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        typenavigator::TimeNavigatorPtr aron(new typenavigator::TimeNavigator(path));
        token->addElement(aron);
    }

    void NavigatorWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }
}
