/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Tuple.h"


namespace armarx::aron::typenavigator
{
    // constructors
    TupleNavigator::TupleNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eTuple, path),
        Navigator(type::Descriptor::eTuple, path),
        type(new type::AronTuple())
    {
    }

    TupleNavigator::TupleNavigator(const type::AronTuplePtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eTuple, path),
        Navigator(type::Descriptor::eTuple, path),
        type(o)
    {
        CheckAronPtrForNull("TupleNavigator", "TupleNavigator", getPath(), o);

        for (const auto& t : type->elementTypes)
        {
            acceptedTypes.push_back(FromAronType(t));
        }
    }

    type::AronTuplePtr TupleNavigator::toAronTuplePtr() const
    {
        if (acceptedTypes.empty())
        {
            throw error::AronException("TupleNavigator", "getCastedResult", "No accepted types set", getPath());
        }
        return type;
    }

    // static methods
    TupleNavigatorPtr TupleNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<TupleNavigator>(n);
    }

    TupleNavigatorPtr TupleNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("TupleNavigator", "DynamicCast[Before]", n);
        TupleNavigatorPtr casted = TupleNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("TupleNavigator", "DynamicCast[After]", n->getPath(), casted);
        return casted;
    }

    // public member functions
    std::vector<NavigatorPtr> TupleNavigator::getAcceptedTypes() const
    {
        return acceptedTypes;
    }

    NavigatorPtr TupleNavigator::getAcceptedType(unsigned int i) const
    {
        return acceptedTypes[i];
    }

    void TupleNavigator::addAcceptedType(const NavigatorPtr& v)
    {
        CheckTypeNavigatorPtrForNull("TupleNavigator", "addAcceptedType", getPath(), v);
        type->elementTypes.push_back(v->getResult());
        acceptedTypes.push_back(v);
    }

    // virtual implementations
    std::vector<NavigatorPtr> TupleNavigator::getChildren() const
    {
        return acceptedTypes;
    }

    size_t TupleNavigator::childrenSize() const
    {
        return acceptedTypes.size();
    }

    type::AronTypePtr TupleNavigator::getResult() const
    {
        return toAronTuplePtr();
    }

    std::string TupleNavigator::getName() const
    {
        std::vector<std::string> names;
        for (const auto& n : acceptedTypes)
        {
            names.push_back(n->getName());
        }
        return "AronTupleType<" + simox::alg::to_string(names, ", ") + ">";
    }
}

