/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "EigenMatrix.h"

namespace armarx::aron::typenavigator
{
    // constructors
    EigenMatrixNavigator::EigenMatrixNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eEigenMatrix, path),
        Navigator(type::Descriptor::eEigenMatrix, path),
        type(new type::AronEigenMatrix({1, 1}, ""))
    {
    }

    EigenMatrixNavigator::EigenMatrixNavigator(const type::AronEigenMatrixPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eDict, path),
        Navigator(type::Descriptor::eEigenMatrix, path),
        type(o)
    {
        CheckAronPtrForNull("AronEigenMatrixNavigator", "AronEigenMatrixNavigator", getPath(), o);
        checkDimensions(type->dimensions);
        checkTypename(type->typeName);
    }

    bool EigenMatrixNavigator::checkDimensions(const std::vector<int>& d) const
    {
        if (d.size() < ACCEPTED_DIMENSION_MIN_SIZE)
        {
            throw error::AronException("AronEigenMatrixNavigator", "checkDimensions", "The dimension size is wrong. Got size: " + std::to_string(d.size()), getPath());
        }

        if (std::any_of(d.begin(), d.end(), [](int i)
    {
        return i < -1 || i == 0;
    }))
        {
            throw error::AronException("AronEigenMatrixNavigator", "checkDimensions", "The dimension size is wrong. At least one empty is < -1 or 0. Got as dimension: " + simox::alg::to_string(d, ", "), getPath());
        }
        return true;
    }

    bool EigenMatrixNavigator::checkTypename(const std::string& s) const
    {
        if (s.empty())
        {
            throw error::AronException("AronEigenMatrixNavigator", "checkTypename", "The typename is empty.", getPath());
        }

        if (!std::any_of(ACCEPTED_TYPES.begin(), ACCEPTED_TYPES.end(), [s](const auto & el)
    {
        return s == el.first || std::find(el.second.begin(), el.second.end(), s) != el.second.end();
        }))
        {
            throw error::AronException("AronEigenMatrixNavigator", "checkTypename", "The typename is wrong. Got " + s, getPath());
        }
        return true;
    }

    unsigned int EigenMatrixNavigator::getRows() const
    {
        return type->dimensions[0];
    }

    unsigned int EigenMatrixNavigator::getCols() const
    {
        return type->dimensions[1];
    }

    void EigenMatrixNavigator::setRows(const unsigned int& w)
    {
        if (w == 0)
        {
            throw error::AronException("AronEigenMatrixNavigator", "setRows", "The rows cannot be 0", getPath());
        }
        type->dimensions[0] = w;
    }

    void EigenMatrixNavigator::setCols(const unsigned int& h)
    {
        if (h == 0)
        {
            throw error::AronException("AronEigenMatrixNavigator", "setCols", "The cols cannot be 0", getPath());
        }
        type->dimensions[1] = h;
    }

    type::AronEigenMatrixPtr EigenMatrixNavigator::toAronEigenMatrixPtr() const
    {
        checkDimensions(type->dimensions);
        checkTypename(type->typeName);
        return type;
    }

    std::string EigenMatrixNavigator::getTypename() const
    {
        return type->typeName;
    }

    std::vector<int> EigenMatrixNavigator::getDimensions() const
    {
        return type->dimensions;
    }

    void EigenMatrixNavigator::setTypename(const std::string& u)
    {
        checkTypename(u);
        type->typeName = u;
    }

    void EigenMatrixNavigator::setDimensions(const std::vector<int>& d)
    {
        checkDimensions(d);
        type->dimensions = d;
    }

    // static methods
    EigenMatrixNavigatorPtr EigenMatrixNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<EigenMatrixNavigator>(n);
    }

    EigenMatrixNavigatorPtr EigenMatrixNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("AronEigenMatrixNavigator", "DynamicCast[Before]", n);
        EigenMatrixNavigatorPtr casted = EigenMatrixNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("AronEigenMatrixNavigator", "DynamicCast[After]", n->getPath(), casted);
        return casted;
    }

    // virtual implementations
    std::vector<NavigatorPtr> EigenMatrixNavigator::getChildren() const
    {
        return {};
    }

    size_t EigenMatrixNavigator::childrenSize() const
    {
        return 0;
    }

    type::AronTypePtr EigenMatrixNavigator::getResult() const
    {
        return toAronEigenMatrixPtr();
    }

    std::string EigenMatrixNavigator::getName() const
    {
        return "AronEigenMatrix<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
    }
}

