/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::aron
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2019
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::aron

#define ARMARX_BOOST_TEST

// STD/STL
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <numeric>

// IVT
#include <Image/ByteImage.h>

// Eigen
#include <Eigen/Core>

// OpenCV
#include <opencv2/core/core.hpp>

// PCL
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>

// Boost
#include <boost/algorithm/string.hpp>

// Test
#include <RobotAPI/Test.h>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>
#include <RobotAPI/libraries/aron/core/Exception.h>

// Aron
#include <RobotAPI/libraries/aron/core/Debug.h>
#include <RobotAPI/libraries/aron/core/Randomizer.h>

#include <RobotAPI/libraries/aron/core/io/dataIO/reader/navigator/NavigatorReader.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/nlohmannJSON/NlohmannJSONReader.h>

#include <RobotAPI/libraries/aron/core/io/dataIO/writer/navigator/NavigatorWriter.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/writer/nlohmannJSON/NlohmannJSONWriter.h>

// Files without code generation
//#include "aronDataWithoutCodeGeneration.h"

// Generated File
#include <RobotAPI/libraries/aron/core/test/aron/NaturalIKTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/ListTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/DictTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PrimitiveTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/ObjectTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/IVTCByteImageTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EigenMatrixTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EigenQuaternionTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/OpenCVMatTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PCLPointCloudTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PositionTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/OrientationTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PoseTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EnumTest.aron.generated.h>

using namespace armarx;
using namespace aron;

template <typename T>
void runTestWithInstances(T& k1, T& k2)
{
    Randomizer r;

    std::cout << "\t getting type 1" << std::endl;
    typenavigator::ObjectNavigatorPtr k1_type_nav = k1.toInitialAronType();
    type::AronObjectPtr k1_type = k1_type_nav->toAronObjectPtr();
    BOOST_CHECK_NE(k1_type.get(), nullptr);

    std::cout << "\t getting type 2" << std::endl;
    typenavigator::ObjectNavigatorPtr k2_type_nav = k2.toInitialAronType();
    type::AronObjectPtr k2_type = k2_type_nav->toAronObjectPtr();
    BOOST_CHECK_NE(k2_type.get(), nullptr);

    std::cout << "\t getting aron 1" << std::endl;
    datanavigator::DictNavigatorPtr k1_aron_nav = k1.toAron();
    data::AronDictPtr k1_aron = k1_aron_nav->toAronDictPtr();
    BOOST_CHECK_NE(k1_aron.get(), nullptr);

    std::cout << "\t initialize aron 1 randomly" << std::endl;
    r.initializeRandomly(k1_aron_nav, k1_type_nav);

    //std::cout << "K Aron:" << std::endl;
    //std::cout << AronDebug::AronDataPtrToString(k_aron) << std::endl;

    std::cout << "\t setting aron 2 from aron 1" << std::endl;
    k2.fromAron(k1_aron);

    std::cout << "\t getting aron 2" << std::endl;
    datanavigator::DictNavigatorPtr k2_aron_nav = k2.toAron();
    data::AronDictPtr k2_aron = k2_aron_nav->toAronDictPtr();
    BOOST_CHECK_NE(k2_aron.get(), nullptr);

    //std::cout << "K2 Aron:" << std::endl;
    //std::cout << AronDebug::AronDataPtrToString(k2_aron) << std::endl;

    std::cout << "\t setting aron 1 from aron 2 and check for equality" << std::endl;

    k1.fromAron(k2_aron_nav);

    BOOST_CHECK_EQUAL((k1 == k2), true);


    //std::cout << "\t getting current type 1" << std::endl;
    //typenavigator::NavigatorPtr k1_current_type_nav = k1.toCurrentAronType();
    //type::AronTypePtr k1_current_type = k1_current_type_nav->getResult();
    //BOOST_CHECK_NE(k1_current_type.get(), nullptr);

    //std::cout << "\t getting type 2" << std::endl;
    //typenavigator::NavigatorPtr k2_current_type_nav = k2.toCurrentAronType();
    //type::AronTypePtr k2_current_type = k2_current_type_nav->getResult();
    //BOOST_CHECK_NE(k2_current_type.get(), nullptr);

    r.initializeRandomly(k1_aron_nav, k1_type_nav);

    k1.fromAron(k1_aron);


    std::cout << "\t check JSON export of k and k2 for equality" << std::endl;
    armarx::aron::dataIO::writer::NlohmannJSONWriter json_writer_for_k1;
    armarx::aron::dataIO::writer::NlohmannJSONWriter json_writer_for_k2;

    k1.write(json_writer_for_k1);
    nlohmann::json k1_aron_json = json_writer_for_k1.getResult();
    std::string k1_aron_json_str = k1_aron_json.dump(4);

    armarx::aron::dataIO::writer::NlohmannJSONWriter direct_json_writer_for_k1;
    armarx::aron::dataIO::Visitor::VisitAndSetup(direct_json_writer_for_k1, k1_aron);
    nlohmann::json direct_k1_aron_json = direct_json_writer_for_k1.getResult();
    std::string direct_k1_aron_json_str = direct_k1_aron_json.dump(4);
    BOOST_CHECK_EQUAL((k1_aron_json_str == direct_k1_aron_json_str), true);

    armarx::aron::dataIO::reader::NlohmannJSONReader json_reader_for_k2(k1_aron_json);
    k2.read(json_reader_for_k2);

    k2.write(json_writer_for_k2);
    nlohmann::json k2_aron_json = json_writer_for_k2.getResult();
    std::string k2_aron_json_str = k2_aron_json.dump(4);

    //std::cout << "\t K1 as json: " << std::endl << k1_aron_json_str << std::endl;
    //std::cout << "\t K2 as json: " << std::endl << k2_aron_json_str << std::endl;
    BOOST_CHECK_EQUAL((k1_aron_json_str == k2_aron_json_str), true);
}


BOOST_AUTO_TEST_CASE(AronNaturalIKTest)
{
    std::cout << "Running NaturalIK test" << std::endl;
    NaturalIKResult k;
    NaturalIKResult k2;
    runTestWithInstances<NaturalIKResult>(k, k2);

    // also test Path
    datanavigator::DictNavigatorPtr aron = k.toAron();
    Path path = aron->getPath(); // should be empty since aron is top level object
    Path memberReached(path, "reached");
    Path memberJointValues(path, "jointValues");
    Path indexJointValues0(memberJointValues, "0");
    Path indexJointValues1(memberJointValues, "1");

    BOOST_CHECK_EQUAL(path.toString(), "\\");
    BOOST_CHECK_EQUAL(path.size(), 0);

    BOOST_CHECK_EQUAL(memberReached.toString(), "\\->reached");
    BOOST_CHECK_EQUAL(memberReached.size(), 1);

    BOOST_CHECK_EQUAL(memberJointValues.toString(), "\\->jointValues");
    BOOST_CHECK_EQUAL(memberJointValues.size(), 1);

    BOOST_CHECK_EQUAL(indexJointValues0.toString(), "\\->jointValues->0");
    BOOST_CHECK_EQUAL(indexJointValues0.size(), 2);

    BOOST_CHECK_EQUAL(indexJointValues1.toString(), "\\->jointValues->1");
    BOOST_CHECK_EQUAL(indexJointValues1.size(), 2);

    datanavigator::BoolNavigatorPtr reached = datanavigator::BoolNavigator::DynamicCastAndCheck(aron->navigateAbsolute(memberReached));
    datanavigator::ListNavigatorPtr jointValues = datanavigator::ListNavigator::DynamicCastAndCheck(aron->navigateAbsolute(memberJointValues));

    if (jointValues->childrenSize() > 0)
    {
        datanavigator::FloatNavigatorPtr el = datanavigator::FloatNavigator::DynamicCastAndCheck(aron->navigateAbsolute(indexJointValues0));
    }
    if (jointValues->childrenSize() > 1)
    {
        datanavigator::FloatNavigatorPtr el = datanavigator::FloatNavigator::DynamicCastAndCheck(aron->navigateAbsolute(indexJointValues1));
    }

    Path diff = indexJointValues1.getWithoutPrefix(indexJointValues0);
    BOOST_CHECK_EQUAL(diff.toString(), "\\->1");
    BOOST_CHECK_EQUAL(diff.size(), 1);
}


BOOST_AUTO_TEST_CASE(AronListTest)
{
    std::cout << "Running List test" << std::endl;
    ListTest l;
    ListTest l2;
    runTestWithInstances<ListTest>(l, l2);
}

BOOST_AUTO_TEST_CASE(AronDictTest)
{
    std::cout << "Running Dict test" << std::endl;
    DictTest d;
    DictTest d2;
    runTestWithInstances<DictTest>(d, d2);
}

BOOST_AUTO_TEST_CASE(AronPrimitiveTest)
{
    std::cout << "Running Primitive test" << std::endl;
    PrimitiveTest p;
    PrimitiveTest p2;
    runTestWithInstances<PrimitiveTest>(p, p2);
}

BOOST_AUTO_TEST_CASE(AronObjectTest)
{
    std::cout << "Running Object test" << std::endl;
    ObjectTest1 o1;
    ObjectTest1 o12;
    runTestWithInstances<ObjectTest1>(o1, o12);

    ObjectTest2 o2;
    ObjectTest2 o22;
    runTestWithInstances<ObjectTest2>(o2, o22);
}

BOOST_AUTO_TEST_CASE(AronImageTest)
{
    std::cout << "Running Image test" << std::endl;
    IVTCByteImageTest i;
    IVTCByteImageTest i2;
    runTestWithInstances<IVTCByteImageTest>(i, i2);
}

/*BOOST_AUTO_TEST_CASE(AronImageChangeSizeTest)
{
    std::cout << "Running Image test with changed sizes" << std::endl;
    IVTCByteImageTest ii;
    IVTCByteImageTest ii2;
    ii.the_ivt_image.Set(100, 100, ii.the_ivt_image.type);
    ii2.the_ivt_image.Set(10, 10, ii.the_ivt_image.type);
    runTestWithInstances<IVTCByteImageTest>(ii, ii2);
}*/

BOOST_AUTO_TEST_CASE(AronEigenMatrixTest)
{
    std::cout << "Running EigenMatrix test" << std::endl;
    EigenMatrixTest em;
    EigenMatrixTest em2;
    runTestWithInstances<EigenMatrixTest>(em, em2);
}

BOOST_AUTO_TEST_CASE(AronOpenCVTest)
{
    std::cout << "Running OpenCVMat test" << std::endl;
    OpenCVMatTest ocv;
    OpenCVMatTest ocv2;
    runTestWithInstances<OpenCVMatTest>(ocv, ocv2);
}

BOOST_AUTO_TEST_CASE(AronPCLPointCloudTest)
{
    std::cout << "Running PCLPointCloud test" << std::endl;
    PointCloudTest pc;
    PointCloudTest pc2;
    runTestWithInstances<PointCloudTest>(pc, pc2);
}

/*
BOOST_AUTO_TEST_CASE(AronTests)
{
    initialize_random();

    std::cout << "Running (Raphael) Trivial test" << std::endl;
    AronTrivialDataWithoutCodeGeneration rt1;
    AronTrivialDataWithoutCodeGeneration rt2;
    runTestWithInstances<AronTrivialDataWithoutCodeGeneration>(rt1, rt2);

    std::cout << "Running (Raphael) Simple test" << std::endl;
    AronSimpleDataWithoutCodeGeneration rs1;
    AronSimpleDataWithoutCodeGeneration rs2;
    runTestWithInstances<AronSimpleDataWithoutCodeGeneration>(rs1, rs2);

    std::cout << "Running (Raphael) Nested test" << std::endl;
    AronNestedDataWithoutCodeGeneration rn1;
    AronNestedDataWithoutCodeGeneration rn2;
    runTestWithInstances<AronNestedDataWithoutCodeGeneration>(rn1, rn2);
}*/

BOOST_AUTO_TEST_CASE(AronEigenQuaternionTest)
{
    std::cout << "Running EigenQuaternion test" << std::endl;
    EigenQuaternionTest eq;
    EigenQuaternionTest eq2;
    runTestWithInstances<EigenQuaternionTest>(eq, eq2);
}

BOOST_AUTO_TEST_CASE(AronPositionTest)
{
    std::cout << "Running Position test" << std::endl;
    PositionTest pc;
    PositionTest pc2;
    runTestWithInstances<PositionTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronOrientationTest)
{
    std::cout << "Running Orientation test" << std::endl;
    OrientationTest pc;
    OrientationTest pc2;
    runTestWithInstances<OrientationTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronPoseTest)
{
    std::cout << "Running Pose test" << std::endl;
    PoseTest pc;
    PoseTest pc2;
    runTestWithInstances<PoseTest>(pc, pc2);
}
